//////////////////////////////////////////////////////////////////////
// sangoma_port.cpp: implementation of the sangoma_port class.
//
// Author	:	David Rokhvarg	<davidr@sangoma.com>
//////////////////////////////////////////////////////////////////////

#include "sangoma_port.h"

#define DBG_CFG		if(0)printf("PORT:");if(0)printf
#define _DBG_CFG	if(0)printf

#define INFO_CFG	if(0)printf("PORT:");if(0)printf
#define _INFO_CFG	if(0)printf

#define ERR_CFG		if(1)printf("PORT:");if(1)printf
#define _ERR_CFG	if(1)printf

#if !defined(__FUNCTION__)
#define	__FUNCTION__ "function"
#endif

#define DEV_NAME_PREFIX	"\\\\.\\"

//////////////////////////////////////////////////////////////////////
//common functions
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

sangoma_port::sangoma_port()
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "sangoma_port";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);
	wp_handle = INVALID_HANDLE_VALUE;
}

sangoma_port::~sangoma_port()
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "~sangoma_port";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	if(wp_handle != INVALID_HANDLE_VALUE){
		CloseHandle(wp_handle);
		wp_handle = INVALID_HANDLE_VALUE;
	}
}



int sangoma_port::init(uint16_t wanpipe_number)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "init";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	wp_number = wanpipe_number;

#if defined(__WINDOWS__)
	_snprintf((char*)wanpipe_name_str, WAN_DRVNAME_SZ, "%sWANPIPE%d", 
		DEV_NAME_PREFIX, wanpipe_number);
#else
	_snprintf((char*)wanpipe_name_str, WAN_DRVNAME_SZ, "/dev/wanpipe");
#endif

	wp_handle = open_api_device((char*)wanpipe_name_str);
	if(wp_handle == INVALID_HANDLE_VALUE){
		ERR_CFG("Error: failed to open %s!!\n", wanpipe_name_str);
		return 1;
	}
	return 0;
}


HANDLE sangoma_port::open_api_device(char *api_device_name)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "open_api_device";
#endif
//	DBG_CFG("%s():%s\n", __FUNCTION__, api_device_name);

#if defined (__WINDOWS__)
	return CreateFile(	api_device_name, 
						GENERIC_READ | GENERIC_WRITE, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						(LPSECURITY_ATTRIBUTES)NULL, 
						OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH,
						(HANDLE)NULL
						);
#else
	int fd=-1;

	char fname[WAN_DRVNAME_SZ];

	sprintf(fname,api_device_name);

	fd = open(fname, O_RDWR);

	return fd;
#endif
}

int sangoma_port::PortManagementCmd(int cmd, void *data, int data_size)
{
	port_management_struct_t port_management;

	memset(&port_management, 0x00, sizeof(port_management_struct_t));

	port_management.command_code = cmd;
	port_management.port_no	= wp_number;
	//some user commands contain data for API driver, copy user's data into API buffer
	if(data_size > 0){
		//copy data, if any
		memcpy(port_management.data, data, data_size);
	}

	if(PortManagementIoctl(&port_management)){
		return SANG_STATUS_INVALID_DEVICE;
	}

	//some commands return data from API driver, copy data into user's buffer
	switch(cmd)
	{
	case GET_HARDWARE_INFO:
		memcpy(data, port_management.data, sizeof(hardware_info_t));
		break;
	case GET_DRIVER_VERSION:
		memcpy(data, port_management.data, sizeof(DRIVER_VERSION));
		break;
	case GET_PORT_OPERATIONAL_STATS:
		memcpy(data, port_management.data, sizeof(port_stats_t));
		break;
	}

	DBG_CFG("%s(): return code: %s\n", __FUNCTION__, SDLA_DECODE_SANG_STATUS(port_management.operation_status));
	return port_management.operation_status;
}


int sangoma_port::PortManagementIoctl(port_management_struct_t *port_management)
{

#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "PortManagementIoctl";
#endif

#if defined(__WINDOWS__)
	DWORD ln;
	if(DeviceIoControl(
			wp_handle,
			IoctlPortManagementCommand,
			(LPVOID)port_management,
			sizeof(port_management_struct_t),
			(LPVOID)port_management,
			sizeof(port_management_struct_t),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		ERR_CFG("%s():Error: IoctlPortManagementCommand failed!!\n", __FUNCTION__);
		return 1;
	}else {
		return 0;
	}
#else
	int err;
	err=ioctl(wp_handle,WANPIPE_IOCTL_PORT_MGMT,port_management);
	if (err) {
		return 1;
	} else {
		return 0;
	}
#endif


}

int sangoma_port::get_hardware_info(hardware_info_t *hardware_info)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "get_hardware_info";
#endif
	return PortManagementCmd(GET_HARDWARE_INFO, hardware_info, 0);
}

int sangoma_port::stop_or_start_port(int command)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "stop_or_start_port";
#endif
	return PortManagementCmd(command, NULL, 0);
}

int sangoma_port::push_a_card_into_wanpipe_info_array(
								wanpipe_instance_info_t *wanpipe_info_array,
								hardware_info_t *new_hw_info)
{
	unsigned int card_ind;
	hardware_info_t *tmp_hw_info;

	//check this card is not in the array already
	for(card_ind = 0; card_ind < MAX_CARDS; card_ind++){
		tmp_hw_info = &wanpipe_info_array[card_ind].hardware_info;

		if(	(new_hw_info->pci_bus_number == tmp_hw_info->pci_bus_number) &&
			(new_hw_info->pci_slot_number == tmp_hw_info->pci_slot_number)){
			//already in array. not an error.
			DBG_CFG("already in array.\n");
			return 1;
		}
	}

	//find an empty slot in the array and place the new card there
	for(card_ind = 0; card_ind < MAX_CARDS; card_ind++){
		tmp_hw_info = &wanpipe_info_array[card_ind].hardware_info;
		if(tmp_hw_info->card_model == -1){
			//found an empty slot in array
			memcpy(tmp_hw_info, new_hw_info, sizeof(hardware_info_t));
			return 0;
		}
	}

	//too many cards in the computer.
	INFO_CFG("Warning: number of cards in the computer is greater than the maximum of %d.\n", 
		MAX_CARDS);
	return 2;
}

//
//Scan all running wanpipes, check which ones are on the same card,
//figure out how many physical cards installed.
//
//Returns:	-1 - if error.
//			number of cards, starting from zero.
int sangoma_port::scan_for_sangoma_cards(wanpipe_instance_info_t *wanpipe_info_array, int card_model)
{
	uint16_t card_counter = 0, wp_ind;
	hardware_info_t	hardware_info;

	//initialize hardware information array.
	memset(wanpipe_info_array, -1, sizeof(wanpipe_instance_info_t)*MAX_CARDS);

	DBG_CFG("%s(): wanpipe_info_array: 0x%p\n", __FUNCTION__, wanpipe_info_array);

	for(wp_ind = 0; wp_ind < (MAX_CARDS	* MAX_PORTS_PER_CARD); wp_ind++){
		if(init(wp_ind)){
			continue;
		}
		if(get_hardware_info(&hardware_info)){
			card_counter = -1;
			break;
		}

		DBG_CFG("card_model		: %s (0x%08X)\n",
			SDLA_ADPTR_NAME(hardware_info.card_model), hardware_info.card_model);
		DBG_CFG("firmware_version\t: 0x%02X\n", hardware_info.firmware_version);
		DBG_CFG("pci_bus_number\t\t: %d\n", hardware_info.pci_bus_number);
		DBG_CFG("pci_slot_number\t\t: %d\n", hardware_info.pci_slot_number);
		DBG_CFG("port_number\t\t: %d\n", hardware_info.port_number);
		DBG_CFG("max_hw_ec_chans\t\t: %d\n", hardware_info.max_hw_ec_chans);
		
		if(card_model == hardware_info.card_model){
			if(push_a_card_into_wanpipe_info_array(wanpipe_info_array, &hardware_info) == 0){
				//At this point we know: 1. WANPIPE number, for example WANPIPE1.
				//						 2. Card type, for example A104.
				//It allows to figure out Device Name of each port on the card.
				//For example: for A104 it will be WANPIPE1, WANPIPE2, WANPIPE3, WANPIPE4
				wanpipe_info_array[card_counter].wanpipe_number = wp_ind;
				DBG_CFG("%s(): wp_ind: %u\n", __FUNCTION__, wp_ind);
				card_counter++;
			}
		}else{
			DBG_CFG("Found a Card but Card Model is different from what was searched for.\n");
		}
	}//for()

	return card_counter;
}

int sangoma_port::get_port_statistics(port_stats_t *port_stats)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "get_port_statistics";
#endif
	return PortManagementCmd(GET_PORT_OPERATIONAL_STATS, port_stats, 0);
}

int sangoma_port::flush_port_statistics()
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "get_port_statistics";
#endif
	return PortManagementCmd(FLUSH_PORT_OPERATIONAL_STATS, NULL, 0);
}
