/*****************************************************************************
* wanpipe_kernel_types.h WANPIPE(tm) - Abstracted Kernel Mode Data structures for
*				Sangoma MS Windows Driver.
*
* Author: 	David Rokhvarg <davidr@sangoma.com
*
* Copyright:	(c) 1984-2008 Sangoma Technologies Inc.
*
* ============================================================================
* Nov 5, 2008  David Rokhvarg	Initial version
* ============================================================================
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
*/

#ifndef _WANPIPE_KERNEL_TYPES_H
#define _WANPIPE_KERNEL_TYPES_H

#if defined(__KERNEL__)

#include "wanpipe_api_iface.h"
#include "wanpipe_api_hdr.h"	/* wp_api_element_t */
#include "wanpipe_api_depricated.h"
#include "wanpipe_cfg_def.h"	/* WAN_IFNAME_SZ */


/***********************************************************************************/
/* upper and lower limits on size of skb buffer: */
#define MAX_SKB_LEN	8188+64 /* 64 is WAN_MAX_HDR_SZ in wanpipe_api_hdr.h. Monitor changes in WAN_MAX_HDR_SZ!! */
#define MIN_SKB_LEN	1

typedef struct sk_buff;


typedef struct sk_buff_head {

    struct sk_buff  *next;
    struct sk_buff  *prev;

    __u32           qlen;
	//maximum allowed number of nodes in the list
	//u32 maximum_size;
}sk_buff_head_t;

struct sk_buff {
	/* sk_buff_head_t must be first! */
	sk_buff_head_t;

	/**********************************************************************/
	/*************** Do NOT change these after allocation *****************/
	u8	*head;		/* Originally allocated buffer pointer. */
	u8	*end;		/* End of originally allocated buffer.  */
	u32 truesize;	/* Length of ORIGINANLLY REQUESTED data buffer - BEFORE adjustments. */
	u32	actual_buffer_len;	/* Length of data buffer AFTER adjustments. */
	/*************** End of "Do NOT change" *****************/
	/**********************************************************************/

	u8	*data;  /* current 1-st byte of user data pointer */
	u8	*tail;  /* current last byte of user data pointer */

	u16	protocol;
	u32	data_len;/* current length of user data */
	u8	cloned;

	void	*list;
	int32	csum;
	u8		*cb;

	void	*dev;

	dma_addr_t	dma_phys;
}sk_buff_t;


struct ifreq{

	u8*  ifr_data;
};

struct seq_file{

	int32 count;
};


#define GFP_ATOMIC	1
#define DBG_NET_DEV 1
#define TMP_STR_LEN		300

#define MAX_SKB_QUEUE_SIZE	100

enum DEV_TYPE {
	DEV_TYPE_LEVEL2_BUS_FDO = 5,
	DEV_TYPE_SDLA_CARD_FDO,
	DEV_TYPE_NET_DEVICE,
	DEV_TYPE_PROTOCOL_FDO,
	DEV_TYPE_CDEV
};

#define DECODE_DEV_TYPE(device_type)			\
	(device_type == DEV_TYPE_LEVEL2_BUS_FDO) ?	\
	"DEV_TYPE_LEVEL2_BUS_FDO"	:				\
	(device_type == DEV_TYPE_SDLA_CARD_FDO)	?	\
	"DEV_TYPE_SDLA_CARD_FDO"	:				\
	(device_type == DEV_TYPE_NET_DEVICE)	?	\
	"DEV_TYPE_NET_DEVICE"	:					\
	(device_type == DEV_TYPE_PROTOCOL_FDO)	?	\
	"DEV_TYPE_PROTOCOL_FDO"	:					\
	(device_type == DEV_TYPE_CDEV)	?	\
	"DEV_TYPE_CDEV"	:					\
	"Unknown device type!!"

/***********************************************************************************/

typedef struct timeval 
{ 
	long_t tv_sec;
	long_t tv_usec;
} timeval;

typedef struct _wan_spinlock
{
	KSPIN_LOCK	slock;
	KIRQL		old_irql;	/* irql BEFORE the spin lock was locked. Use ONLY for non-IRQ locks!! */
	PKINTERRUPT	pInterruptObject;
	u32			magic_number;/* allows to check if spinlock was initialized */
	long_t		is_locked;/* current state of spinlock: 0-unlocked, 1-locked. 
						InterlockedIncrement() and InterlockedDecrement() require PLONG. */
#define SPINLOCK_NAME_LEN	64
	char		name[SPINLOCK_NAME_LEN];
}wan_spinlock_t;

typedef KIRQL		wan_smp_flag_t;

typedef struct _wan_taskq
{
	unsigned char		running;
	KDPC 	tqueue;
} wan_taskq_t;

typedef void*				wan_timer_arg_t;
typedef PKDEFERRED_ROUTINE	wan_tasklet_func_t;
typedef PKDEFERRED_ROUTINE	wan_taskq_func_t;

typedef struct _wan_timer_info_t{
	KTIMER			Timer;
	KDPC			TimerDpcObject;
	LARGE_INTEGER		TimerDueTime;
}wan_timer_info_t;

typedef void (*wan_timer_func_t)(IN PKDPC Dpc, void* context, void * arg2, void * arg3);

/*
** TIMER structure
*/
typedef struct _wan_timer
{
#define NDIS_TIMER_TAG 0xBEEF0005
	unsigned long		Tag;
	wan_timer_func_t    MiniportTimerFunction;
	void *              MiniportTimerContext;
	void *              MiniportAdapterHandle;
	wan_timer_info_t	timer_info;
	wan_timer_func_t	timer_func;
	void*				timer_arg;
} wan_timer_t;

/***********************************************************************************/

typedef struct _user_buffer_ifo
{
	void			*queue/*DATA_QUEUE*/;	//Allocated on driver startup. Used ONLY at hardware level.
	PIRP			pendingIrp;
    PUCHAR			userbfr;
	ULONG			userbfr_len;

	FAST_MUTEX		PendingIrpFastMutex;
	KEVENT			WaitEvent;
#define UI_BUFFER_NAME_LEN 64
	char			ui_buffer_name[UI_BUFFER_NAME_LEN];

	KSPIN_LOCK		spinlock_irp;//used ONLY in sprotocol.sys driver!

}user_buffer_info_t;

/***********************************************************************************/

typedef struct _sdla_wanpipe_interface
{
	//pointer to the wanpipe interface, passed when calling 'receive_indicate()'
	void* wanpipe_if;
	//pointer to receive function, called by VirIfEnum.sys
	void (*receive_indicate)(void* wanpipe_if, wp_api_element_t *rx_data);

	//to indicate to network layer we've completed TX and we can TX again.
	//called by sdladrv.sys
	void (*tx_complete_indicate)(void* wanpipe_if);

	//pointer to 'netdevice_t', passed when calling 'transmit_indicate()'
	void* p_netdevice;
	//pointer to transmit function, called by sprotocol.sys
	int (*transmit_indicate)(void* p_netdevice, wp_api_element_t *tx_data);

	//pointer to Management input routine of underlying layer
	int (*management_indicate)(	void* p_called_netdevice, void *wan_udp/*'wan_udp_hdr_t'*/);

	//pointer to Management Complete routine. Called by underlying layer
	//to indicate completion of Management call.
	int (*management_complete)(	void* p_netdevice);

	/*to indicate to layer above about: 
		1. T1/E1 Connect/Disconnect events.
		other events can be added
	*/
	int (*out_of_band_event_indicate)(void* p_netdevice, unsigned int event_type, void *optional_event_data);

	//this name used by wanpipe.sys for passing it to wanpipemon,
	//when used on a Network interface.
	unsigned char	underlying_dev_name[WAN_IFNAME_SZ];

}sdla_wanpipe_interface_t;


typedef struct _NOTIFY_RECORD{

	PKEVENT			UserEvent;

} NOTIFY_RECORD, *PNOTIFY_RECORD;

enum _FLAGS_BIT_MAP{
	IFF_UP
};

struct wanif_conf;

typedef struct net_device
{
	u8 device_type; // must be first member of the structure!!

	u8 flags;
	u32 mtu;
	u32 type; 
	u32 addr_len; 	/* ETH_ALEN;*/
	u32 hard_header_len;
	u32 tx_queue_len;
	u32 irq;
	u32 dma;
	u32 base_addr;

	u32 mem_start; 
	u32 mem_end;
	ulong_t trans_start;
	u32 watchdog_timeo;

	u8 name[WAN_IFNAME_SZ];

	u8	tbusy;//used by wpabs_netif_queue_stopped()

	void *priv;//pointer to the private area

	void *hard_header;
	void *rebuild_header;

	int		(*init)(void* dev);
	int		(*open)(void* dev);
	int		(*stop)(void* dev);
	int		(*hard_start_xmit)(void* skb, void* dev);
	void*	(*get_stats)(void* dev);
	void	(*tx_timeout) (void* netdev);
	int		(*do_ioctl)(void* dev, void* ifr, int cmd);
	int		(*udp_mgmt)(void* card, void* net_dev, void* priv_area, int not_used);
	int		(*change_mtu)(void *dev, int new_mtu);

	struct wanif_conf *wanif_conf;	//configuration of the logical interface.
	int		i_encapsulation_type;	//one of: WPLIP_RAW, WPLIP_IP

	//////////////////////////////////////////////////////////////
	//Rx Data related members
	user_buffer_info_t rx_user_info;
	//////////////////////////////////////////////////////////////
	//Tx related members
	user_buffer_info_t tx_user_info;
	//////////////////////////////////////////////////////////////
	//API Poll related members
	user_buffer_info_t api_poll_user_info;
	//////////////////////////////////////////////////////////////
	//UDP and IOCTL Management (WANPIPEMON) related stuff
	user_buffer_info_t mgmt_user_info;
	//////////////////////////////////////////////////////////////

	void			*sdla_card;			//pointer back to 'parent' sdla card

	UNICODE_STRING	SymbolicLinkString;	//device name accessible from user mode. NOT used in sdladrv.sys.
	PDEVICE_OBJECT	fdo;	//NOT used in sdladrv.sys.

	int	open_handle_counter;

	void* lip_dev;	//pointer to LIP layer interface (i.g. DLCI). used to pass data for transmission.

	//////////////////////////////////////////////////////////////
	//DMA stuff
	PDMA_ADAPTER		AdapterObject;				// DMA adapter object
	ULONG				nMapRegisters;				// maximum # mapping registers
	PHYSICAL_ADDRESS	txPhysicalLogicalAddress;
	PHYSICAL_ADDRESS	rxPhysicalLogicalAddress;

	void* rx_dma_buf;
	void* tx_dma_buf;

	///////////////////////////////////////////////////////////////
	//initialization flag
	char init_done;

	///////////////////////////////////////////////////////////////
	//these needed to pass received data to driver above:
	//
	//pointer to the wanpipe interface, passed when calling 'receive_indicate()'
	void* wanpipe_if;
	//pointer to receive function, called by sdladrv.sys 
	void (*receive_indicate)(void* card, wp_api_element_t* rx_data);

	//to indicate to network layer we've completed TX and we can TX again.
	void (*tx_complete_indicate)(void* card);

	//to indicate to layer above we've completed a Management call.
	//Note about parameter: when called by sdladrv.sys it is 'card',
	//						when called by sprotocol it is 'p_netdevice'
	//						and usually accepted by wanpipe.sys (NDIS driver),
	//						but does not have to (can be next level of protocols).
	int (*management_complete)(	void *card_or_p_netdevice);

	wp_api_element_t rx_data;	//Temp buffer for Wanpipe above this interface.
							//used in both sdladrv.sys and sprotocol.sys.
							//Also used by TDMV API for temporary storage
							//for RX Codec operations.

	//wp_tdm_chan_stats_t	if_stats;

	void* trace_info;//points to wan_trace_t in chan pointed by 'priv'

	/* to indicate to layer above about: 
		1. T1/E1 Connect/Disconnect events.
		other events can be added */
	int (*out_of_band_event_indicate)(void* card, u32 event_type, void *optional_event_data);


	///////////////////////////////////////////////////////////////
	u8 (*set_tx_idle_data_in_priv)(void *p_netdevice);

	u32 interface_number;	//zero-based number. used by A200 analog code to find out the 'line number'

//	void			*wp_tdm_api_dev_ptr;

	//this will be set at each level on Line connect/disconnecte events
	u32	current_line_state;//T1/E1 connected/disconnected

	u8	usedby;//API/STACK/LIB SANGOMA API...

	PNOTIFY_RECORD	rx_data_shared_event_record;

	u8	hdlc_eng;

	wan_taskq_t	sdla_netdev_task;
	u32			sdla_netdev_task_bits;

	struct sk_buff *pending_tx_skb;

}netdevice_t;

#define NETDEV_MGMT_UI(sdla_net_dev)	(sdla_net_dev->mgmt_user_info)
#define NETDEV_POLL_UI(sdla_net_dev)	(sdla_net_dev->api_poll_user_info)
#define NETDEV_RX_UI(sdla_net_dev)		(sdla_net_dev->rx_user_info)
#define NETDEV_TX_UI(sdla_net_dev)		(sdla_net_dev->tx_user_info)

#define NETDEV_MGMT_UI_MUTEX(sdla_net_dev)	(NETDEV_MGMT_UI(sdla_net_dev).PendingIrpFastMutex)
#define NETDEV_POLL_UI_MUTEX(sdla_net_dev)	(NETDEV_POLL_UI(sdla_net_dev).PendingIrpFastMutex)
#define NETDEV_RX_UI_MUTEX(sdla_net_dev)	(NETDEV_RX_UI(sdla_net_dev).PendingIrpFastMutex)
#define NETDEV_TX_UI_MUTEX(sdla_net_dev)	(NETDEV_TX_UI(sdla_net_dev).PendingIrpFastMutex)

#define NETDEV_MGMT_UI_WAIT_EVENT(sdla_net_dev)	(NETDEV_MGMT_UI(sdla_net_dev).WaitEvent)
#define NETDEV_POLL_UI_WAIT_EVENT(sdla_net_dev)	(NETDEV_POLL_UI(sdla_net_dev).WaitEvent)
#define NETDEV_RX_UI_WAIT_EVENT(sdla_net_dev)	(NETDEV_RX_UI(sdla_net_dev).WaitEvent)
#define NETDEV_TX_UI_WAIT_EVENT(sdla_net_dev)	(NETDEV_TX_UI(sdla_net_dev).WaitEvent)

#define NETDEV_RX_UI_DATA_QUEUE(sdla_net_dev)	(NETDEV_RX_UI(sdla_net_dev).queue)
#define NETDEV_TX_UI_DATA_QUEUE(sdla_net_dev)	(NETDEV_TX_UI(sdla_net_dev).queue)

#define NETDEV_INIT_UI_BUFFER(ui_buff, ui_buff_name)	{					\
	ExInitializeFastMutex(&ui_buff.PendingIrpFastMutex);					\
	KeInitializeEvent(&ui_buff.WaitEvent, NotificationEvent, FALSE);		\
	_snprintf(ui_buff.ui_buffer_name, UI_BUFFER_NAME_LEN, ui_buff_name);	\
	KeInitializeSpinLock(&ui_buff.spinlock_irp);							\
}

#define NETDEV_SET_WAIT_EVENT(WaitEvent)	{			\
	KeSetEvent (&WaitEvent, IO_NO_INCREMENT, FALSE);	\
	KeClearEvent(&WaitEvent);							\
}

/************* IRQ macros ************/
typedef BOOLEAN	irqreturn_t;

#define IRQ_NONE        FALSE
#define IRQ_HANDLED     TRUE

#define IRQ_RETVAL(x)   ((x) != IRQ_NONE)

#define WAN_IRQ_CALL(fn,args,ret)	ret = fn args
#define WAN_IRQ_RETURN(a) 			return a
#define WAN_IRQ_RETVAL				irqreturn_t
#define WAN_IRQ_RETVAL_DECL(ret)	irqreturn_t ret = WAN_IRQ_NONE
#define WAN_IRQ_RETVAL_SET(ret, val)	ret = val
#define WAN_IRQ_HANDLED 			IRQ_HANDLED
#define WAN_IRQ_NONE	 			IRQ_NONE

#define WAN_IRQ_DECODE(irqrc)				\
(irqrc == IRQ_HANDLED)	? "IRQ_HANDLED"	:	\
(irqrc == IRQ_NONE)		? "IRQ_NONE"	:	\
"Invalid IRQ RC"

//////////////////////////////////////////////////////////////////////////////////////////////

typedef struct aft_dma_chain
{
	ulong_t	init;
	u32		dma_addr;
	u32		dma_len;
//	struct sk_buff 	*skb;
	u32		index;

	u32		dma_descr;
	u32		len_align;
	u32		reg;

	void *dma_buf;

}aft_dma_chain_t;

//////////////////////////////////////////////////////////////////////////////////////////////

#define DECODE_NT_STATUS(status)												\
	status == STATUS_SUCCESS ? "STATUS_SUCCESS":								\
	status == STATUS_UNSUCCESSFUL ? "STATUS_UNSUCCESSFUL":						\
	status == STATUS_INSUFFICIENT_RESOURCES ? "STATUS_INSUFFICIENT_RESOURCES":	\
	status == STATUS_INVALID_PARAMETER ? "STATUS_INVALID_PARAMETER":			\
	status == STATUS_OBJECT_NAME_NOT_FOUND ? "STATUS_OBJECT_NAME_NOT_FOUND":	\
	status == STATUS_OBJECT_NAME_INVALID ? "STATUS_OBJECT_NAME_INVALID":		\
	status == STATUS_OBJECT_NAME_COLLISION ? "STATUS_OBJECT_NAME_COLLISION":	\
	"Unknown NT Status"

//////////////////////////////////////////////////////////////////////////////////////////////

#endif	//WAN_KERNEL
#endif	//_WANPIPE_KERNEL_TYPES_H


